function N(X) { // Number of cases
	return X.length;
}
function mean(X) { // Calculates the mean
	return sigmaX(X) / X.length; 
}
function mode(X) { // Calculates the mode
// See the following for objects that count frequencies 
// https://stackoverflow.com/questions/5667888/counting-the-occurrences-frequency-of-array-elements
// See also conditional (ternary) operator https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Conditional_Operator
// See also the getKeyByValue function below for searching object by values to get keys.
	var counts = {}; // An object to hold the values and their associated frequencies
	var values = []; // An array for the object keys (data without duplicate entries)
	var highest = 0; // For finding the most frequently occuring number
	for (var i = 0; i < X.length; i++) { // An object that counts the frequency of each value in the array.
		var num = X[i];
		counts[num] = counts[num] ? counts[num] + 1 : 1; // Uses brief format for if-then
	}
	var values = Object.keys(counts); // This creates an array of the keys
	for (i = 0; i < values.length; i++) { 
		if (highest < counts[values[i]]) {var highest = counts[values[i]];} // finds the highest frequency value
	}
//	console.log(counts); for debugging
//	console.log(values);
//	console.log(highest);
	return getKeyByValue(counts, highest); // Searches the object counts by highest to return the highest frequency values
}
function maximum(X) { // Find the largest value in the array
//See https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Math/max
	return Math.max(...X);
}
function minimum(X) { // Find the smallest value in the array
	return Math.min(...X);
}
function sortAscending(X) { // Sort from lowest to highest values
//See https://www.w3schools.com/jsref/jsref_sort.asp
//The compare function is important for sorting numbers. The default is to sort alphabetically.
	return X.sort(function(a, b){return a - b});
}
function sortDescending(X) { // Sort from highest to lowest values
	return X.sort(function(a, b){return b - a});
}
function median(X) { // Calculate the median
//See https://www.w3resource.com/javascript-exercises/fundamental/javascript-fundamental-exercise-88.php
//The modulus (%) operator is for remainder and is used for distinguishing odd vs. even numbers.
	sortedX = sortAscending(X);
	midpoint = Math.floor(sortedX.length / 2); // The middle position is array length divided by 2.
//	console.log("Test for odd values, where 1 = true: ",X.length % 2);
//	console.log("Array position midpoint:",midpoint);
	if (X.length % 2 == 1) {return sortedX[midpoint]} // For odd numbers in the array choose the middle value
	// For even numbered arrays. The midpoint - 1 is important because array numbering starts at 0
		else {return (sortedX[midpoint-1] + sortedX[midpoint]) / 2}
}
function rangeExc(X) {  // Range exclusive is highest minus lowest. Commonly used. 
	return maximum(X) - minimum(X);
}
function rangeInc(X) {  // Range inclusive is less commonly used than exclusive range. 
	return maximum(X) - minimum(X) + 1;
}
function sigmaX(X) { //Calculates SigmaX
	var sum = 0;
		for (i = 0; i < X.length; i++) {sum = sum + (X[i]);}
	return sum;
}
function sigmaX2(X) { //Calculates SigmaX^2
	var sum = 0;
		for (i = 0; i < X.length; i++) {sum = sum + Math.pow(X[i],2);}
	return sum;
}
function sumsSquares(X) {  // Sums of squares for variance and standard deviation
	var sum = 0;
	var meanX = mean(X); 
	for (i = 0; i < X.length; i++) {sum = sum + Math.pow(X[i]-meanX,2);}
	return sum
}
function sigmaXY (X,Y) { // Sigma XY for correlation, regression
	if (X.length != Y.length) {return "Warning: X and Y have an unequal number of values.";}
	var sum = 0;
	for (i = 0; i < X.length; i++) {sum = sum + (X[i] * Y[i]);}
	return sum;
}
function sigmaD (X1,X2) { // Difference scores for dependent-samples t-test
	if (X1.length != X2.length) {return "Warning: X and Y have an unequal number of values.";}
	var sum = 0; 
	for (i = 0; i < X1.length; i++) {sum = sum + (X2[i] - X1[i]);} // Sums difference values for post minus pre 
	return sum; 
}
function sigmaD2 (X1,X2) { // Squared difference scores for dependent-samples t-test
	if (X1.length != X2.length) {return "Warning: X and Y have an unequal number of values.";}
	var sum = 0; 
	for (i = 0; i < N(X1); i++) {sum = sum + Math.pow((X2[i] - X1[i]),2);} // Sums the squared difference values (post minus pre)
	return sum; 
}
function varianceBiased (X) { // Variance for a population, division by N. 
	return sumsSquares(X) / N(X); 
}
function varianceUnbiased (X) {
	return sumsSquares(X) / (N(X) - 1); // Variance for a sample; division by N - 1
}
function SDBiased (X) {
	return Math.sqrt(varianceBiased(X));  // SD for a population
}
function SDUnbiased (X) {
	return Math.sqrt(varianceUnbiased(X)); // SD for a sample
}
function pearsonr (X,Y) {
	if (X.length != Y.length) {return "Warning: X and Y have an unequal number of values.";}
	return ((N(X)*sigmaXY(X,Y)) - (sigmaX(X)*sigmaX(Y))) / Math.sqrt((N(X) * sigmaX2(X) - Math.pow(sigmaX(X),2))*(N(Y) * sigmaX2(Y) - Math.pow(sigmaX(Y),2)));
}
function regressionSlope (X,Y) {
	if (X.length != Y.length) {return "Warning: X and Y have an unequal number of values.";}
	return (sigmaXY(X,Y) - (sigmaX(X)*sigmaX(Y)/N(X))) / (sigmaX2(X) - (Math.pow(sigmaX(X),2) / N(X)));
}
function regressionYInt (X,Y) {
	if (X.length != Y.length) {return "Warning: X and Y have an unequal number of values.";}
	return (sigmaX(Y) - (regressionSlope(X,Y) * sigmaX(X))) / N(X);
}
function oneSampleZTest (X,Y,Z) {
	return (mean(X) - Y) / (Z / Math.sqrt(N(X))); 
}
function indSamplesTTest (X,Y) { 
	return (mean(X) - mean(Y)) / Math.sqrt(((N(X)-1)*varianceUnbiased(X) + (N(Y)-1)*varianceUnbiased(Y))/((N(X)+N(Y)-2)/((N(X)+N(Y))/(N(X)*N(Y))))); 
}
function relatedSamplesTTest (X,Y) {
	return sigmaD(X,Y) / Math.sqrt((N(X) * sigmaD2(X,Y) - Math.pow(sigmaD(X,Y),2))/(N(X) - 1));
}
function oneSampleTTest (X,Y) {
	return (mean(X) - Y) / (SDUnbiased(X) / Math.sqrt(N(X)));
}
function cohensD (X,Y) {
// formula from Salkind & Frey (2020), p. 208
	return (mean(X) - mean(Y)) / Math.sqrt((varianceUnbiased(X)+varianceUnbiased(Y))/2); 
}
function cohensDDescriptive (M1,M2,SD1,SD2) {
// formula from Salkind and Frey (2020), p. 208, with modification for standard deviation input
	return (M1 - M2) / Math.sqrt((Math.pow(SD1,2) + Math.pow(SD2,2)) / 2);
}
function cohensDav (X1,X2) {
// For dependent-samples t-tests
// formula from https://real-statistics.com/students-t-distribution/paired-sample-t-test/cohens-d-paired-samples/
// Configured for post (X2) minus pre (X1), unlike cohensD above
// This is the same formula as cohensD above. 
	return (mean(X2) - mean(X1)) / Math.sqrt((varianceUnbiased(X2)+varianceUnbiased(X1))/2); 
}
// ANOVA sources: Heiman (2006), p. 315-320; Salkind & Frey, 
function SStotal (X,Y,Z) {
	var sigmaXtotal = sigmaX(X) + sigmaX(Y) + sigmaX(Z);
	var sigmaXsquaredtotal = sigmaX2(X) + sigmaX2(Y) + sigmaX2(Z);
	var Ntotal = N(X) + N(Y) + N(Z);
	var SStot = sigmaXsquaredtotal - (Math.pow(sigmaXtotal,2)/Ntotal); // SS total 
	return SStot; 
}
function SSbetween (X,Y,Z) {
	var sigmaXtotal = sigmaX(X) + sigmaX(Y) + sigmaX(Z);
	var Ntotal = N(X) + N(Y) + N(Z);
	var SSbtwn = (Math.pow(sigmaX(X),2)/N(X)) + (Math.pow(sigmaX(Y),2)/N(Y)) + (Math.pow(sigmaX(Z),2)/(N(Z))) - (Math.pow(sigmaXtotal,2)/Ntotal);
	return SSbtwn; 
}
function SSwithin (X,Y,Z) {
	var SSwn = SStotal(X,Y,Z) - SSbetween(X,Y,Z); 
	return SSwn;
}
function MSbetween (X,Y,Z) {
	return SSbetween(X,Y,Z)/2; // dfbetweeen = 3 groups - 1 = 2  
}
function MSwithin (X,Y,Z) {
	var dfwtn = N(X) + N(Y) + N(Z) - 3; // degrees of freedom within, -3 is for number of groups 
	return SSwithin(X,Y,Z)/dfwtn;
}
function ANOVA (X,Y,Z) {
	return MSbetween(X,Y,Z)/MSwithin(X,Y,Z);
}
function EtaSquare (X,Y,Z) {
	return SSbetween(X,Y,Z) / SStotal(X,Y,Z); 
}
function EtaSquaredFromSS (btwnSS,totalSS) {
	return btwnSS/totalSS; 
}
function chiSqGoodnessFit (X) {
	var expected = sigmaX(X) / X.length // Expected = sum of frequencies / number of groups
	var sum = 0; 
	for (i = 0; i < X.length; i++) {var sum = sum + (Math.pow(X[i]-expected,2)/expected);}
	return sum; 
}
function chiSqGoodnessExpected (X) {
	return expected = sigmaX(X) / X.length // Expected = sum of frequencies / number of groups
}
function chiSqGoodnessPercent (X) {
	var out = ""; // An output string variable
	var total = sigmaX(X);
	for (i = 0; i < X.length; i++) {var out = out + Math.round((X[i]/total)*10000)/100 + "% ";} // Percentages rounded
	return out; 
}
function chiSqInd (X,Y) {
	var row1Total = sigmaX(X);  // row 1 total
	var row2Total = sigmaX(Y);  // row 2 total
	var overallTotal = sigmaX(X) + sigmaX(Y);  // Overall N
	var expectedRow1 = []; // Arrays to hold the expected values
	var expectedRow2 = [];
	for (i = 0; i < X.length; i++) { // Compute expected values
		expectedRow1.push((row1Total * (X[i]+Y[i])) / overallTotal); // df = (row total + column total) / overall total
		expectedRow2.push((row2Total * (X[i]+Y[i])) / overallTotal); 
		}
		console.log(expectedRow1,expectedRow2);
	var sum = 0; 
	for (i = 0; i < X.length; i++) {  // Calculate the chi-square independence test
		sum = sum + (Math.pow(X[i]-expectedRow1[i],2)/expectedRow1[i]);
		sum = sum + (Math.pow(Y[i]-expectedRow2[i],2)/expectedRow2[i]);
	}
	return sum; 
}
function chiSqIndExpected (X,Y) {
	var out = ""; // An output string variable
	var row1Total = sigmaX(X); // row 1 total
	var row2Total = sigmaX(Y); // row 2 total
	var overallTotal = sigmaX(X) + sigmaX(Y);  // Overall N
	for (i = 0; i < X.length; i++) {var out = out + ((row1Total * (X[i]+Y[i])) / overallTotal) + ", ";}
	var out = out + "<br/>Row 2 E = "; // Space between rows
	for (i = 0; i < Y.length; i++) {var out = out + ((row2Total * (X[i]+Y[i])) / overallTotal) + ", ";} 
	return out;
}
function chiSqIndPercent (X,Y) {
	var out = ""; // An output string variable
	var total = sigmaX(X) + sigmaX(Y);  // Overall N
	for (i = 0; i < X.length; i++) {var out = out + Math.round((X[i]/total)*10000)/100 + "% ";} // Percentages rounded
	var out = out + "<br/>Row 2: "; // Space between rows
	for (i = 0; i < Y.length; i++) {var out = out + Math.round((Y[i]/total)*10000)/100 + "% ";} // Percentages rounded
	return out;
}
//*** The following is for inter-quartile range. The midpoint seems correct. Dividing the array into a lower part needs work.
//function IQRQ1 (X) {
//	var data = sortAscending(X);
//	var midpoint = Math.floor(data.length/2) - 1;
//	if (X.length % 2 == 1) {var midpoint = Math.floor(data.length/2) - 1;} // Midpoint, rounded down and -1 for odd scores
//		else {var midpoint = Math.floor(data.length/2);} // Midpoint for even scores 
//	for (i = midpoint; i < data.length; i++) {data.pop(); console.log(i);}
//	console.log("midpoint: ",midpoint);
//	console.log(data);
//	return median(data);
//}
// The following function is to search an object by value to get a particular key. Needed for mode. 
// See https://stackoverflow.com/questions/9907419/how-to-get-a-key-in-a-javascript-object-by-its-value
// Modification for .filter to return multiple values in case of bimodal, trimodal, etc. situations.
// The .find method only returns the first match.
// See https://www.digitalocean.com/community/tutorials/js-array-search-methods
function getKeyByValue(object, value) {
  return Object.keys(object).filter(key => object[key] === value);
}